{ TAudioVolume

   This component controls volume level or muting state of audio device and application.

   Any suggestions/advice would be appreciated.

   Author : Silhwan Hyun  (hyunsh@hanafos.com) }



unit AudioVolumeKol;

interface

uses
  Kol, Windows, Messages, ActiveKOL,
  share, objects, MMSystem, ActiveX;
const
// * Added constants to specify device add/removal event
  DEVICE_STATE_ADDED = $80000000;
  DEVICE_STATE_REMOVED = $80000001;
const
  CLASS_IMMDeviceEnumerator             : TGUID = '{BCDE0395-E52F-467C-8E3D-C4579291692E}';
  IID_IMMDeviceEnumerator               : TGUID = '{A95664D2-9614-4F35-A746-DE8DB63617E6}';
  IID_IMMDevice                         : TGUID = '{D666063F-1587-4E43-81F1-B948E807363F}';
  IID_IMMDeviceCollection               : TGUID = '{0BD7A1BE-7A1A-44DB-8397-CC5392387B5E}';
  IID_IAudioEndpointVolume              : TGUID = '{5CDF2C82-841E-4546-9722-0CF74078229A}';
  IID_IAudioMeterInformation            : TGUID = '{C02216F6-8C67-4B5B-9D00-D008E73E0064}';
  IID_IAudioEndpointVolumeCallback      : TGUID = '{657804FA-D6AD-4496-8A60-352752AF4F89}';

  IID_IAudioClient                      : TGUID = '{1CB9AD4C-DBFA-4c32-B178-C2F568A703B2}';
  IID_IMMNotificationClient             : TGUID = '{7991EEC9-7E89-4D85-8390-6C703CEC60C0}';
  IID_IAudioSessionEvents               : TGUID = '{24918ACC-64B3-37C1-8CA9-74A66E9957A8}';
  IID_IAudioSessionControl              : TGUID = '{F4B1A599-7266-4319-A8CA-E70ACB11E8CD}';
  IID_IAudioSessionControl2             : TGUID = '{bfb7ff88-7239-4fc9-8fa2-07c950be9c6d}';  // >= Windows 7
  IID_ISimpleAudioVolume                : TGUID = '{87CE5498-68D6-44E5-9215-6DA47EF883D8}';
  IID_IAudioSessionManager              : TGUID = '{BFA971F1-4D5E-40BB-935E-967039BFBEE4}';
  IID_IAudioSessionManager2             : TGUID = '{77AA99A0-1BD6-484F-8BC7-2C654C9A9B6F}';  // >= Windows 7
  IID_IAudioCaptureClient               : TGUID = '{C8ADBD64-E71E-48a0-A4DE-185C395CD317}';
  IID_IAudioRenderClient                : TGUID = '{F294ACFC-3146-4483-A7BF-ADDCA7C260E2}';

  DEVICE_STATE_ACTIVE                   = $00000001;
  DEVICE_STATE_DISABLED                 = $00000002;    // **  Added (19 Mar 2012)
  DEVICE_STATE_NOTPRESENT               = $00000004;
  DEVICE_STATE_UNPLUGGED                = $00000008;    // **  Fixed (incorrectly defined, 19 Mar 2012)
  DEVICE_STATEMASK_ALL                  = $0000000F;    // **  Fixed (incorrectly defined, 19 Mar 2012)

  eRender                               = $00000000;
  eCapture                              = $00000001;
  eAll                                  = $00000002;
  EDataFlow_enum_count                  = $00000003;

  eConsole                              = $00000000;
  eMultimedia                           = $00000001;
  eCommunications                       = $00000002;
  ERole_enum_count                      = $00000003;

  ENDPOINT_HARDWARE_SUPPORT_VOLUME    = $00000001;
  ENDPOINT_HARDWARE_SUPPORT_MUTE      = $00000002;
  ENDPOINT_HARDWARE_SUPPORT_METER     = $00000004;

  AUDCLNT_SHAREMODE_SHARED              = 0;
  AUDCLNT_SHAREMODE_EXCLUSIVE           = 1;

  AUDCLNT_BUFFERFLAGS_DATA_DISCONTINUITY   = 1;
  AUDCLNT_BUFFERFLAGS_SILENT               = 2;
  AUDCLNT_BUFFERFLAGS_TIMESTAMP_ERROR      = 4;

  AUDCLNT_STREAMFLAGS_LOOPBACK        = $00020000;
  AUDCLNT_STREAMFLAGS_EVENTCALLBACK   = $00040000;
  AUDCLNT_STREAMFLAGS_NOPERSIST       = $00080000;

 // Constants for AudioSessionState value
  AudioSessionStateInactive = 0;       // The last running stream in the session stops.
  AudioSessionStateActive = 1;         // The session has one or more streams that are running.
  AudioSessionStateExpired = 2;        // The client destroys the last stream in the session by releasing all
                                       // references to the stream object.
 // Constants for DisconnectReason value
  DisconnectReasonDeviceRemoval = 0;   // The user removed the audio endpoint device.
  DisconnectReasonServerShutdown = 1;  // The Windows audio service has stopped.
  DisconnectReasonFormatChanged = 2;   // The stream format changed for the device that the audio session is connected to.
  DisconnectReasonSessionLogoff = 3;   // The user logged off the Windows Terminal Services (WTS) session that the audio
                                       // session was running in.
  DisconnectReasonSessionDisconnected = 4; // The WTS session that the audio session was running in was disconnected.
  DisconnectReasonExclusiveModeOverride = 5;  // The (shared-mode) audio session was disconnected to make the audio endpoint
                                              // device available for an exclusive-mode connection.

type
   IInitializeWithFile = interface(IUnknown)
   ['{b7d14566-0509-4cce-a71f-0a554233bd9b}']
     function Initialize(pszFilePath: PAnsiChar; grfMode: DWORD):HRESULT; stdcall;
   end;


   IInitializeWithStream = interface(IUnknown)
   ['{b824b49d-22ac-4161-ac8a-9916e8fa3f7f}']
     function Initialize(var pIStream: IStream; grfMode: DWORD):HRESULT; stdcall;
   end;

   _tagpropertykey = packed record
     fmtid: TGUID;
     pid: DWORD;
   end;
   PROPERTYKEY = _tagpropertykey;
   PPropertyKey = ^TPropertyKey;
   TPropertyKey = _tagpropertykey;

   IPropertyStore = interface(IUnknown)
     ['{886d8eeb-8cf2-4446-8d02-cdba1dbdcf99}']
     function GetCount(out cProps: DWORD): HResult; stdcall;
     function GetAt(iProp: DWORD; out pkey: TPropertyKey): HResult; stdcall;
     function GetValue(const key: TPropertyKey; out pv: TPropVariant):HResult; stdcall;
     function SetValue(const key: TPropertyKey; const propvar:TPropVariant): HResult; stdcall;
     function Commit: HResult; stdcall;
   end;

   IPropertyStoreCapabilities = interface(IUnknown)
   ['{c8e2d566-186e-4d49-bf41-6909ead56acc}']
     function IsPropertyWritable(pPropKey: PPropertyKey): HRESULT; stdcall;
   end;

const
   PKEY_Device_FriendlyName: TPropertyKey = (
                fmtid: (D1:$a45c254e; D2:$df1c ; D3:$4efd;
                D4: ($80, $20, $67, $d1, $46, $a8, $50, $e0));
                pid: 14);

type
 // {$if CompilerVersion < 20} // if compiler is older version than Delphi 2009
  UINT32 = integer;
  PUINT64 = ^int64;
 // {$ifend}
  REFERENCE_TIME = int64;
  EDataFlow = TOleEnum;
  ERole = TOleEnum;
  AUDCLNT_SHAREMODE = TOleEnum;

  PAUDIO_VOLUME_NOTIFICATION_DATA = ^AUDIO_VOLUME_NOTIFICATION_DATA;
  AUDIO_VOLUME_NOTIFICATION_DATA = packed record
    guidEventContext: TGUID;
    bMuted: BOOL;
    fMasterVolume: Single;
    nChannels: UINT;
    afChannelVolumes: array [1..1] of Single;
  end;

type
  IAudioEndpointVolumeCallback = interface(IUnknown)
  ['{657804FA-D6AD-4496-8A60-352752AF4F89}']
    function OnNotify(pNotify:PAUDIO_VOLUME_NOTIFICATION_DATA):HRESULT; stdcall;
  end;

  IAudioEndpointVolume = interface(IUnknown)
    ['{5CDF2C82-841E-4546-9722-0CF74078229A}']
    function RegisterControlChangeNotify(pNotify: IAudioEndpointVolumeCallback): HRESULT; stdcall;
    function UnregisterControlChangeNotify(pNotify: IAudioEndpointVolumeCallback): HRESULT; stdcall;
    function GetChannelCount(var pnChannelCount: UINT): HRESULT; stdcall;
    function SetMasterVolumeLevel(fLevelDB: Single; pguidEventContext: PGuid): HRESULT; stdcall;
    function SetMasterVolumeLevelScalar(fLevel: Single; pguidEventContext: PGuid): HRESULT; stdcall;
    function GetMasterVolumeLevel(var pfLevelDB: Single): HRESULT; stdcall;
    function GetMasterVolumeLevelScalar(var pfLevel: Single): HRESULT; stdcall;
    function SetChannelVolumeLevel(nChannel: UINT; fLevelDB: Single; pguidEventContext: PGuid): HRESULT; stdcall;
    function SetChannelVolumeLevelScalar(nChannel: UINT; fLevel:Single; pguidEventContext: PGuid): HRESULT; stdcall;
    function GetChannelVolumeLevel(nChannel: UINT; fLevelDB: Single): HRESULT; stdcall;
    function GetChannelVolumeLevelScalar(nChannel: UINT; fLevel: Single): HRESULT; stdcall;
    function SetMute(bMute: BOOL; pguidEventContext: PGuid): HRESULT; stdcall;
    function GetMute(var pbMute: BOOL): HRESULT; stdcall;
    function GetVolumeStepInfo(var pnStep: UINT; var pnStepCount: UINT): HRESULT; stdcall;
    function VolumeStepUp(pguidEventContext: PGuid): HRESULT; stdcall;
    function VolumeStepDown(pguidEventContext: PGuid): HRESULT; stdcall;
    function QueryHardwareSupport(var pdwHardwareSupportMask: UINT): HRESULT; stdcall;
    function GetVolumeRange(var pflVolumeMindB: Single; var pflVolumeMaxdB: Single;
                            var pflVolumeIncrementdB: Single): HRESULT; stdcall;
  end;

  IAudioMeterInformation = interface(IUnknown)
  ['{C02216F6-8C67-4B5B-9D00-D008E73E0064}']
    function GetPeakValue(out fPeak: Single): HRESULT; stdcall;
    function GetMeteringChannelCount(out nChannelCount: UINT): HRESULT; stdcall;
    function GetChannelsPeakValues(u32ChannelCount: UINT; pfPeakValues: pSingle): HRESULT; stdcall;
    function QueryHardwareSupport(out dwHardwareSupportMask: UINT): HRESULT; stdcall;
  end;

  IMMDevice = interface(IUnknown)
    ['{D666063F-1587-4E43-81F1-B948E807363F}']
    function Activate(const iid: TGUID; dwClsCtx: UINT; pActivationParams: PPropVariant; out ppInterface:IUnknown): HRESULT; stdcall;
    function OpenPropertyStore(stgmAccess: integer; out ppProperties: IPropertyStore): HRESULT; stdcall;  // ** Changed at 2011-08-09
    function GetId(out ppstrId: PWChar): HRESULT; stdcall;
    function GetState(var pdwState: Integer): HRESULT; stdcall;
  end;

  IMMDeviceCollection = interface(IUnknown)
    ['{0BD7A1BE-7A1A-44DB-8397-CC5392387B5E}']
    function GetCount(var pcDevices: UINT):HRESULT; stdcall;
    function Item(nDevice: UINT; out ppDevice: IMMDevice):HRESULT; stdcall;
  end;

  IMMNotificationClient = interface(IUnknown)
  ['{7991EEC9-7E89-4D85-8390-6C703CEC60C0}']
   function OnDeviceStateChanged(DeviceId: LPCWSTR; NewState: DWORD): HRESULT; stdcall;
   function OnDeviceAdded(DeviceId: LPCWSTR): HRESULT; stdcall;
   function OnDeviceRemoved(DeviceId: LPCWSTR): HRESULT; stdcall;
   function OnDefaultDeviceChanged(Flow: EDataFlow; Role: ERole; NewDefaultDeviceId: LPCWSTR): HRESULT; stdcall;
   function OnPropertyValueChanged(DeviceId: LPCWSTR; const Key: PROPERTYKEY): HRESULT; stdcall;
  end;

  IMMDeviceEnumerator = interface(IUnknown)
    ['{A95664D2-9614-4F35-A746-DE8DB63617E6}']
    function EnumAudioEndpoints(dataFlow:EDataFlow; dwStateMask:DWORD; out ppDevices: IMMDeviceCollection): HRESULT; stdcall;
    function GetDefaultAudioEndpoint(dataFlow:EDataFlow; role: ERole; out ppEndpoint: IMMDevice): HRESULT; stdcall;
    function GetDevice(pwstrId: PWChar;out ppDevice:IMMDevice): HRESULT; stdcall;
  // **  Fixed following 2 functions (parameter pClient was incorrectly defined as var, 19 Mar 2012)
    function RegisterEndpointNotificationCallback(const pClient: IMMNotificationClient): HRESULT; stdcall;
    function UnregisterEndpointNotificationCallback(const pClient: IMMNotificationClient): HRESULT; stdcall;
  end;

  IAudioClient = interface(IUnknown)
  ['{1CB9AD4C-DBFA-4c32-B178-C2F568A703B2}']
   function Initialize(ShareMode: AUDCLNT_SHAREMODE; StreamFlags: DWORD; hmsBufferDuration: REFERENCE_TIME;
    hmsPeriodicity: REFERENCE_TIME; pFormat: PWAVEFORMATEX; AudioSessionGuid: pGuid): HRESULT; stdcall;
   function GetBufferSize(out NumBufferFrames: UINT32): HRESULT; stdcall;
   function GetStreamLatency(out hmsLatency: REFERENCE_TIME): HRESULT; stdcall;
   function GetCurrentPadding(out NumPaddingFrames: UINT32): HRESULT; stdcall;
   function IsFormatSupported(ShareMode: DWORD; pFormat: PWAVEFORMATEX; out pClosestMatch: PWAVEFORMATEX): HRESULT; stdcall;
   function GetMixFormat(out pFormat: PWAVEFORMATEX): HRESULT; stdcall;
   function GetDevicePeriod(out hmsDefaultDevicePeriod, hmsMinimumDevicePeriod: REFERENCE_TIME): HRESULT; stdcall;
   function Start: HRESULT; stdcall;
   function Stop: HRESULT; stdcall;
   function Reset: HRESULT; stdcall;
   function SetEventHandle(eventHandle: HWND): HRESULT; stdcall;
   function GetService(const iid: TGUID; out ppInterface:IUnknown): HRESULT; stdcall;
  end;

  IAudioSessionEvents = interface(IUnknown)
  ['{24918ACC-64B3-37C1-8CA9-74A66E9957A8}']
   function OnDisplayNameChanged(NewDisplayName: LPCWSTR; EventContext: pGuid): HResult; stdcall;
   function OnIconPathChanged(NewIconPath: LPCWSTR; EventContext: pGuid): HResult; stdcall;
   function OnSimpleVolumeChanged(NewVolume: Single; NewMute: LongBool; EventContext: pGuid): HResult; stdcall;
   function OnChannelVolumeChanged(ChannelCount: uint; NewChannelArray: PSingle; ChangedChannel: uint;
    EventContext: pGuid): HResult; stdcall;
   function OnGroupingParamChanged(NewGroupingParam, EventContext: pGuid): HResult; stdcall;
   function OnStateChanged(NewState: uint): HResult; stdcall; // AudioSessionState
   function OnSessionDisconnected(DisconnectReason: uint): HResult; stdcall; // AudioSessionDisconnectReason
  end;

  IAudioSessionControl = interface(IUnknown)
  ['{F4B1A599-7266-4319-A8CA-E70ACB11E8CD}']
   function GetState(out pRetVal: uint): HResult; stdcall;
   function GetDisplayName(out pRetVal: LPWSTR): HResult; stdcall; // pRetVal must be freed by CoTaskMemFree
   function SetDisplayName(Value: LPCWSTR; EventContext: pGuid): HResult; stdcall;
   function GetIconPath(out pRetVal: LPWSTR): HResult; stdcall; // pRetVal must be freed by CoTaskMemFree
   function SetIconPath(Value: LPCWSTR; EventContext: pGuid): HResult; stdcall;
   function GetGroupingParam(pRetVal: pGuid): HResult; stdcall;
   function SetGroupingParam(OverrideValue, EventContext: pGuid): HResult; stdcall;
   function RegisterAudioSessionNotification(const NewNotifications: IAudioSessionEvents): HResult; stdcall;
   function UnregisterAudioSessionNotification(const NewNotifications: IAudioSessionEvents): HResult; stdcall;
  end;

  //Interface IAudioVolumeDuckNotification >= Windows 7
  IAudioVolumeDuckNotification = interface(IUnknown)
  ['{C3B284D4-6D39-4359-B3CF-B56DDB3BB39C}']
    function OnVolumeDuckNotification(const sessionID: LPCWSTR; const countCommunicationSessions: UINT32): HResult; stdcall;
    function OnVolumeUnduckNotification(const sessionID: LPCWSTR): HResult; stdcall;
  end;

  //Interface IAudioSessionControl2 >= Windows 7
  IAudioSessionControl2 = interface(IAudioSessionControl)
  ['{bfb7ff88-7239-4fc9-8fa2-07c950be9c6d}']
    function GetSessionIdentifier(out pRetVal: LPWSTR): Hresult; stdcall;
    function GetSessionInstanceIdentifier(out pRetVal: LPWSTR): Hresult; stdcall;
    function GetProcessId(out pRetVal: DWord): Hresult; stdcall;
    function IsSystemSoundsSession(): Hresult; stdcall;   //indicates whether the session is a system sounds session.
                                                          //S_OK = The session is a system sounds session.
                                                          //S_FALSE = The session is not a system sounds session.
    function SetDuckingPreference(const optOut: Boolean): Hresult; stdcall; //A BOOL variable that enables or disables system auto-ducking.
  end;

  ISimpleAudioVolume = interface(IUnknown)
  ['{87CE5498-68D6-44E5-9215-6DA47EF883D8}']
   function SetMasterVolume(fLevel: Single; EventContext: pGuid): HResult; stdcall;
   function GetMasterVolume(out fLevel: Single): HResult; stdcall;
   function SetMute(bMute: LongBool; EventContext: pGuid): HResult; stdcall;
   function GetMute(out bMute: LongBool): HResult; stdcall;
  end;

  IAudioSessionManager = interface(IUnknown)
  ['{BFA971F1-4D5E-40BB-935E-967039BFBEE4}']
   function GetAudioSessionControl(AudioSessionGuid: pGUID; CrossProcessSession: LongBool;
    out SessionControl: IAudioSessionControl): HResult; stdcall;
   function GetSimpleAudioVolume(AudioSessionGuid: pGuid; StreamFlag: uint;
    out AudioVolume: ISimpleAudioVolume): HResult; stdcall;
  end;

  //Interface IAudioSessionNotification
        IAudioSessionNotification = interface(IUnknown)
  ['{641DD20B-4D41-49CC-ABA3-174B9477BB08}']
    function OnSessionCreated(const NewSession: IAudioSessionControl): HResult; stdcall;
  end;

  //Interface IAudioSessionEnumerator
  IAudioSessionEnumerator = interface(IUnknown)
  ['{E2F5BB11-0570-40CA-ACDD-3AA01277DEE8}']
    function GetCount(out SessionCount: integer): HResult; stdcall;
    function GetSession(const SessionCount: integer; out Session: IAudioSessionControl): HResult; stdcall;
  end;

  //Interface IAudioSessionManager2 >= Windows 7
  IAudioSessionManager2 = interface(IAudioSessionManager)
  ['{77AA99A0-1BD6-484F-8BC7-2C654C9A9B6F}']
    function GetSessionEnumerator(out SessionEnum: IAudioSessionEnumerator): HResult; stdcall;
    function RegisterSessionNotification(SessionNotification: IAudioSessionNotification): HResult; stdcall;
    function UnregisterSessionNotification(SessionNotification: IAudioSessionNotification): HResult; stdcall;
    function RegisterDuckNotification(const sessionID: LPCWSTR; const duckNotification: IAudioVolumeDuckNotification): HResult; stdcall;
    function UnregisterDuckNotification(const duckNotification: IAudioVolumeDuckNotification): HResult; stdcall;
  end;

  IAudioCaptureClient = interface(IUnknown)
  ['{C8ADBD64-E71E-48a0-A4DE-185C395CD317}']
   function GetBuffer(out pData: PBYTE; out NumFramesToRead: UINT32; out dwFlags: DWORD;
    pu64DevicePosition: PUINT64; pu64QPCPosition: PUINT64): HResult; stdcall;
   function ReleaseBuffer(NumFramesRead: UINT32): HResult; stdcall;
   function GetNextPacketSize(out NumFramesInNextPacket: UINT32): HResult; stdcall;
  end;

  IAudioRenderClient = interface(IUnknown)
  ['{F294ACFC-3146-4483-A7BF-ADDCA7C260E2}']
   function GetBuffer(NumFramesRequested: UINT32; out pData: pByte): HResult; stdcall;
   function ReleaseBuffer(NumFramesWritten: UINT32; dwFlags: DWORD): HResult; stdcall;
  end;

var
  Win32MajorVersion: Integer = 0;
  Win32MinorVersion: Integer = 0;
  Win32BuildNumber: Integer = 0;
  Win32Platform: Integer = 0;
  Win32CSDVersion: string = '';

type
  PVolMute = ^TVolMute;
  TVolMute = record
    Volume: single;
    Muted: boolean;
  end;

  TAudioSessionEvents = class(TInterfacedObject, IAudioSessionEvents)
  private
   MsgHandle: HWND;
   VolMute: TVolMute;
  public
  // the functions required to implement
   function OnDisplayNameChanged(NewDisplayName: LPCWSTR; EventContext: pGuid): HResult; stdcall;
   function OnIconPathChanged(NewIconPath: LPCWSTR; EventContext: pGuid): HResult; stdcall;
   function OnSimpleVolumeChanged(NewVolume: Single; NewMute: LongBool; EventContext: pGuid): HResult; stdcall;
   function OnChannelVolumeChanged(ChannelCount: uint; NewChannelArray: PSingle; ChangedChannel: uint;
    EventContext: pGuid): HResult; stdcall;
   function OnGroupingParamChanged(NewGroupingParam, EventContext: pGuid): HResult; stdcall;
   function OnStateChanged(NewState: uint): HResult; stdcall; // AudioSessionState
   function OnSessionDisconnected(DisconnectReason: uint): HResult; stdcall; // AudioSessionDisconnectReason
   procedure SetMsgHandle(WinHandle: HWND);
  end;

  TAudioEndpointEvents = class(TInterfacedObject, IAudioEndpointVolumeCallback)
  private
   MsgHandle: HWND;
   VolMute: TVolMute;
   EventContext: TGUID;
  public
   function OnNotify(pNotify: PAUDIO_VOLUME_NOTIFICATION_DATA):HRESULT; stdcall;
   procedure SetMsgHandle(WinHandle: HWND);
  end;

  TMMNotificationClient = class(TInterfacedObject, IMMNotificationClient)
  private
   MsgHandle: HWND;
   tmpDefaultId: WideString;
  public
   function OnDeviceStateChanged(DeviceId: LPCWSTR; NewState: DWORD): HRESULT; stdcall;
   function OnDeviceAdded(DeviceId: LPCWSTR): HRESULT; stdcall;
   function OnDeviceRemoved(DeviceId: LPCWSTR): HRESULT; stdcall;
   function OnDefaultDeviceChanged(Flow: EDataFlow; Role: ERole; NewDefaultDeviceId: LPCWSTR): HRESULT; stdcall;
   function OnPropertyValueChanged(DeviceId: LPCWSTR; const Key: PROPERTYKEY): HRESULT; stdcall;
   procedure SetMsgHandle(WinHandle: HWND);
  end;

  TDeviceInfo = record     // ** Added at v0.9.4
    DeviceName: WideString;//string;
    DeviceId: WideString; //string;
  end;

// ByMyAction is used to check if this event is triggered by self action, i.e., my application's
// execution of SetMasterMute, SetMasterVolume, SetMute or SetVolume.
  TOnVolumeEvent = procedure(ByMyAction: Boolean; Volume: Single; Mute: Boolean) of object;
  TOnSessionStateEvent = procedure(NewState: integer) of object;
  TOnDeviceStateChange = procedure(DeviceId: string; NewDeviceState: DWORD) of object;
  TOnDefaultDeviceChange = procedure(NewDefaultDevice: TDeviceInfo) of object;
  TOnError = procedure(Err: string) of object;

  TAudioVolume = class(TObject)
  protected
    DeviceEnumerator: IMMDeviceEnumerator;
    DeviceCollection: IMMDeviceCollection;
    AudioEndpointVolume: IAudioEndpointVolume;
    AudioEndpointEvents: IAudioEndpointVolumeCallback;
    MMNotificationClient: TMMNotificationClient;

    SessionEnumerator: IAudioSessionEnumerator;
    SessionControl: IAudioSessionControl;
    SessionEvents: IAudioSessionEvents;
    SimpleAudioVolume: ISimpleAudioVolume;
    PeakMeter: IAudioMeterInformation;
    MasterMeter: IAudioMeterInformation;

    MyVolSet: Boolean;
    MyVolVal: Single;
    MyMuteSet: Boolean;
    MyMuteVal: Boolean;

    FReady: Boolean;
    FDeviceCount: Cardinal;
    FMixFormat: TWAVEFORMATEX;
    FChannels: DWORD;
    FSelectedDevice: TDeviceInfo;
    FDefaultDevice: TDeviceInfo;
    FDeviceList: array of TDeviceInfo;
   // FDeviceNames: array of WideString;
    FEventHandle: HWND;
    FSuppressEvent: Boolean;
    FOnVolumeEvent: TOnVolumeEvent;
    FOnMasterVolumeEvent: TOnVolumeEvent;
    FOnSessionStateEvent: TOnSessionStateEvent;
    FOnDeviceStateChange: TOnDeviceStateChange;
    FOnDefaultDeviceChange: TOnDefaultDeviceChange;
    FOnError: TOnError;

  private
    procedure ProcessMsg(var Msg: TMessage);
    procedure SetSuppressEvent(Value: Boolean);
    function SetupDefaultAudioEndpoint(Device_Enumerator: IMMDeviceEnumerator): Boolean;
    function RefreshDeviceList(Device_Enumerator: IMMDeviceEnumerator): boolean;
    function GetDeviceInfo(Device: IMMDevice): TDeviceInfo;
    function GetDeviceInfo2(DeviceCollection_: IMMDeviceCollection; DeviceIndex: Longword): TDeviceInfo;
    function GetDeviceName(DeviceCollection_: IMMDeviceCollection; DeviceIndex: Longword): WideString;
  public
    constructor Create(AOwner: TObject);
    destructor Destroy; override;

  // function AudioDeviceName(DeviceIndex: Cardinal): WideString;
    function AudioDeviceInfo(DeviceIndex: Cardinal): TDeviceInfo;
    function SelectAudioDevice(DeviceIndex: Cardinal): Boolean;
    property SelectedDevice: TDeviceInfo read FSelectedDevice;
    function IsMasterMuted: Boolean;
    procedure SetMasterMute(Value: Boolean);
    function GetMasterVolume: Single;
    procedure SetMasterVolume(Value: Single); // Value : 0 ~ 1.0
    function IsMuted: Boolean;
    procedure SetMute(Value: Boolean);
    function GetVolume: Single;
    procedure SetVolume(Value: Single);  // Value : 0 ~ 1.0
    function GetMasterPeak: Single;      // Gets Endpoint device's audio output level
    function SetupPeakMeter: boolean;    // ** New at v0.9.5
   // ** Note : function "SetupPeakMeter" should be called after initialization of application.
    function GetPeak: Single;   // Gets application's audio output level
    property DeviceMixFormat: TWAVEFORMATEX read FMixFormat;
    property MasterChannels: DWORD read FChannels;
  //published
    property Ready: Boolean read FReady;       // true if it's ready to operate
    property AudioDeviceCount: Cardinal read FDeviceCount;
    property DefaultAudioDevice: TDeviceInfo read FDefaultDevice;
    property SuppressEvent: Boolean read FSuppressEvent write SetSuppressEvent;
    property OnMasterVolumeEvent: TOnVolumeEvent read FOnMasterVolumeEvent write FOnMasterVolumeEvent;
    property OnVolumeEvent: TOnVolumeEvent read FOnVolumeEvent write FOnVolumeEvent;
    property OnSessionStateEvent: TOnSessionStateEvent read FOnSessionStateEvent write FOnSessionStateEvent;
    property OnDeviceStateChange: TOnDeviceStateChange read FOnDeviceStateChange write FOnDeviceStateChange;
    property OnDefaultDeviceChange: TOnDefaultDeviceChange read FOnDefaultDeviceChange write FOnDefaultDeviceChange;
    property OnError: TOnError read FOnError write FOnError;
  end;

  function PropVariantClear(pvar: PPropVariant): HRESULT; stdcall; external 'ole32.dll';
  procedure PropVariantInit(var pv : TPropVariant);


implementation

const
  MsgBaseOffset = 100;
  WM_EndpointVolume = WM_User + MsgBaseOffset;
  WM_VolumeEvent = WM_User + MsgBaseOffset + 1;
  WM_SessionStateEvent = WM_User + MsgBaseOffset + 2;
  WM_DeviceStateChange = WM_User + MsgBaseOffset + 3;
  WM_DefaultDeviceChange = WM_User + MsgBaseOffset + 4;
  WM_DeviceAdded = WM_User + MsgBaseOffset + 5;
  WM_DeviceRemoved = WM_User + MsgBaseOffset + 6;

  GUID_TAudioVolume: TGUID = '{444C70B1-5C25-4800-A37F-A11223A34037}';

  CyclePerSec = 25;
  REFTIMES_PER_SEC: double = 10000000;
  REFTIMES_PER_MILLISEC = 10000;

procedure PropVariantInit(var pv : TPropVariant);
begin
  pv.vt := VT_EMPTY;
end;

{$IFNDEF UNICODE}
procedure StrCopyW(dest, source: pWideChar; MaxLength: DWORD);
var
  p: pbyte;
  i, k: DWORD;
  b1, b2: byte;
begin
  p := pointer(source);
  i := 0;
  k := MaxLength * 2;

  repeat
    b1 := p^;
    inc(p); inc(i);
    b2 := p^;
    inc(p); inc(i);
    if (b1 = 0) and (b2 = 0) then
      break;
  until i = k;

  CopyMemory(dest, pointer(source), i);
end;
{$ENDIF}

// function GetDeviceInfo retrieves the device friendly name & device id for a particular device.
function TAudioVolume.GetDeviceInfo(Device: IMMDevice): TDeviceInfo;
var
  deviceId: pWideChar;
  propertyStore: IPropertyStore;
  friendlyName: PROPVARIANT;
  deviceNameBuf: array[0..127] of WideChar;
  deviceName: pWideChar;
  HR: HRESULT;
begin
  Result.DeviceName := '';
  Result.DeviceId := '';
  deviceId := nil;

  try
    HR := device.GetId(deviceId);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get device id of a device !');
    Result.DeviceId := WideString(deviceId);

    HR := device.OpenPropertyStore(STGM_READ, propertyStore);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to open device property store of a device !');

    PropVariantInit(friendlyName);
    HR := propertyStore.GetValue(PKEY_Device_FriendlyName, friendlyName);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to retrieve friendly name for a device !');

    if friendlyName.vt = VT_LPWSTR then begin
      deviceName := @deviceNameBuf[0];
      {$IFDEF UNICODE}
      StrCopy(deviceName, friendlyName.pwszVal);
      {$ELSE}
      StrCopyW(deviceName, friendlyName.pwszVal, 127);
      {$ENDIF}
      Result.DeviceName := deviceName;
    end;

  finally
    if deviceId <> nil then
      CoTaskMemFree(deviceId);
    PropVariantClear(@friendlyName);
  end;
end;

// function GetDeviceInfo2 retrieves the device friendly name & device id for a particular device in a device collection.
function TAudioVolume.GetDeviceInfo2(DeviceCollection_: IMMDeviceCollection; DeviceIndex: Longword): TDeviceInfo;
var
  device: IMMDevice;
  HR: HRESULT;
begin
  Result.DeviceName := '';
  Result.DeviceId := '';

  try
    HR := DeviceCollection_.Item(DeviceIndex, device);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get IMMDevice interface for device #' + int2Str(DeviceIndex));

    Result := GetDeviceInfo(device);
  finally

  end;
end;

//-------------------------------------------------------------------------------

function TAudioSessionEvents.OnDisplayNameChanged(NewDisplayName: LPCWSTR; EventContext: pGuid): HResult; stdcall;
begin
  result := S_OK;
end;

function TAudioSessionEvents.OnIconPathChanged(NewIconPath: LPCWSTR; EventContext: pGuid): HResult; stdcall;
begin
  result := S_OK;
end;

function TAudioSessionEvents.OnSimpleVolumeChanged(NewVolume: Single; NewMute: LongBool; EventContext: pGuid): HResult; stdcall;
 // NewVolume : The new volume level for the audio session.
 // NewMute : The new muting state. If TRUE, muting is enabled. If FALSE, muting is disabled.
 // EventContext : The event context value. This is the same value that the caller passed to ISimpleAudioVolume::SetMasterVolume or
 //                ISimpleAudioVolume::SetMute in the call that changed the volume level or muting state of the session
begin
  VolMute.Volume := NewVolume;
  VolMute.Muted := NewMute;
  PostMessage(MsgHandle, WM_VolumeEvent, integer(@VolMute), integer(EventContext));
  // if Assigned(OnError) then FOnError('OnSimpleVolumeChanged');
  result := S_OK;
end;

function TAudioSessionEvents.OnChannelVolumeChanged(ChannelCount: uint; NewChannelArray: PSingle; ChangedChannel: uint;
 EventContext: pGuid): HResult; stdcall;
 // ChannelCount : the number of audio channels in the session submix
 // NewChannelArray : Pointer to an array of volume levels
 // ChangedChannel : The number of the channel whose volume level changed. (- 1  -> more than one channel have changed)
 // EventContext : The event context value. This is the same value that the caller passed to the IChannelAudioVolume::SetChannelVolume or
 //                IChannelAudioVolume::SetAllVolumes method in the call that initiated the change in volume level of the channel.
begin
  result := S_OK;
end;

function TAudioSessionEvents.OnGroupingParamChanged(NewGroupingParam, EventContext: pGuid): HResult; stdcall;
begin
  result := S_OK;
end;

function TAudioSessionEvents.OnStateChanged(NewState: uint): HResult; stdcall;
begin
  PostMessage(MsgHandle, WM_SessionStateEvent, NewState, 0);

  result := S_OK;
end;

function TAudioSessionEvents.OnSessionDisconnected(DisconnectReason: uint): HResult; stdcall;
begin
  result := S_OK;
end;

procedure TAudioSessionEvents.SetMsgHandle(WinHandle: HWND);
begin
  MsgHandle := WinHandle;
end;


function TAudioEndpointEvents.OnNotify(pNotify: PAUDIO_VOLUME_NOTIFICATION_DATA):HRESULT; stdcall;
var
  VolumeData: AUDIO_VOLUME_NOTIFICATION_DATA;
begin
  VolumeData := pNotify^;
  VolMute.Volume := VolumeData.fMasterVolume;
  VolMute.Muted := VolumeData.bMuted;
  EventContext := VolumeData.guidEventContext;
  PostMessage(MsgHandle, WM_EndpointVolume, integer(@VolMute), integer(@EventContext));

  result := S_OK;
end;

procedure TAudioEndpointEvents.SetMsgHandle(WinHandle: HWND);
begin
  MsgHandle := WinHandle;
end;


function TMMNotificationClient.OnDeviceStateChanged(DeviceId: LPCWSTR; NewState: DWORD): HRESULT; stdcall;
begin
  PostMessage(MsgHandle, WM_DeviceStateChange, WPARAM(DeviceId), LPARAM(NewState));

  result := S_OK;
end;

function TMMNotificationClient.OnDeviceAdded(DeviceId: LPCWSTR): HRESULT; stdcall;
begin
  PostMessage(MsgHandle, WM_DeviceAdded, WPARAM(DeviceId), 0);

  result := S_OK;
end;

function TMMNotificationClient.OnDeviceRemoved(DeviceId: LPCWSTR): HRESULT; stdcall;
begin
  PostMessage(MsgHandle, WM_DeviceRemoved, WPARAM(DeviceId), 0);

  result := S_OK;
end;

function TMMNotificationClient.OnDefaultDeviceChanged(Flow: EDataFlow; Role: ERole; NewDefaultDeviceId: LPCWSTR): HRESULT; stdcall;
var
  pw: PWideChar;
begin
  if (Flow = eRender) and (Role = eMultimedia) then
  begin
  // the content pointed by NewDefaultDeviceId is unavailable after exiting this function
    pw := NewDefaultDeviceId;
    tmpDefaultId := WideString(pw);
    pw := PWideChar(tmpDefaultId);
    PostMessage(MsgHandle, WM_DefaultDeviceChange, WPARAM(pw), 0);
  end;

  result := S_OK;
end;

function TMMNotificationClient.OnPropertyValueChanged(DeviceId: LPCWSTR; const Key: PROPERTYKEY): HRESULT; stdcall;
begin
  result := S_OK;
end;

procedure TMMNotificationClient.SetMsgHandle(WinHandle: HWND);
begin
  MsgHandle := WinHandle;
end;
//--------------------------------------------------------------------------------------------


// function GetDeviceName retrieves the device friendly name for a particular device in a device collection.
function TAudioVolume.GetDeviceName(DeviceCollection_: IMMDeviceCollection; DeviceIndex: Longword): WideString;
var
  device: IMMDevice;
  deviceId: pWideChar;
  propertyStore: IPropertyStore;
  friendlyName: PROPVARIANT;
  deviceNameBuf: array[0..127] of WideChar;
  deviceName: pWideChar;
  HR: HRESULT;
begin
  Result := '';
  deviceId := nil;

  try
    HR := DeviceCollection_.Item(DeviceIndex, device);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get IMMDevice interface for device #' + int2Str(DeviceIndex));

    HR := device.GetId(deviceId);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get device id for device #' + int2Str(DeviceIndex));

    HR := device.OpenPropertyStore(STGM_READ, propertyStore);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to open device property store for device #' + int2Str(DeviceIndex));

    PropVariantInit(friendlyName);
    HR := propertyStore.GetValue(PKEY_Device_FriendlyName, friendlyName);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to retrieve friendly name for device #' + int2Str(DeviceIndex));

    if friendlyName.vt = VT_LPWSTR then begin
      deviceName := @deviceNameBuf[0];
      {$IFDEF UNICODE}
      StrCopy(deviceName, friendlyName.pwszVal);
      {$ELSE}
      StrCopyW(deviceName, friendlyName.pwszVal, 127);
      {$ENDIF}
      Result := deviceName;
    end;

  finally
    if deviceId <> nil then
      CoTaskMemFree(deviceId);
    PropVariantClear(@friendlyName);
  end;
end;

// ** function SetupPeakMeter should be called after initialization of application.
function TAudioVolume.SetupPeakMeter: boolean;
var
  HR: HRESULT;
  SessionCount: integer;
  Session: IAudioSessionControl;
  SessionControl2: IAudioSessionControl2;
  I: integer;
 // pSessionName: pWideChar;
 // SessionName: WideString;
  CurProcId, ProcId: Cardinal;
begin
  Result := false;

 // Device := nil;
  PeakMeter := nil;
  if SessionEnumerator = nil then
    exit;

  try
  // iterate over all the sessions
    SessionCount := 0;
    HR := SessionEnumerator.GetCount(SessionCount);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get the number of Audio Sessions');

  // ActiveSessions := 0;
    CurProcId := GetCurrentProcessId;
    for I := 0 to SessionCount - 1 do
    begin
    // Get Audio Session
      HR := SessionEnumerator.GetSession(I, Session);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get the Audio Session');

    // Get Audio Session Control2
      HR := Session.QueryInterface(IID_IAudioSessionControl2, SessionControl2);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get the Session Control2');

      HR := SessionControl2.IsSystemSoundsSession;
      if HR = S_OK then
        continue;

      HR := SessionControl2.GetProcessId(ProcId);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get Process Identifier');

      if ProcId = CurProcId then
      begin
      // Get Audio Session Name
       { HR := SessionControl2.GetSessionIdentifier(pSessionName);
        if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get the Session Name');
        SessionName := pSessionName; }

      // Get Audio Meter Information
        HR := Session.QueryInterface(IID_IAudioMeterInformation, IUnknown(PeakMeter));
        if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get the Audio Meter Information');

        Result := true;
        break;
      end;
    end;

  finally

  end;
end;

function TAudioVolume.SetupDefaultAudioEndpoint(Device_Enumerator: IMMDeviceEnumerator): Boolean;
var
  HR: HRESULT;
  Device: IMMDevice;
  SessionManager: IAudioSessionManager;
  SessionManager2: IAudioSessionManager2;

begin
  Result := false;
  Device := nil;

  try
  // Get default audio endpoint device with the specified data-flow direction (eRender or eCapture)
  //  and device role.
    HR := Device_Enumerator.GetDefaultAudioEndpoint(eRender, eMultimedia, Device);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get DefaultAudioEndpoint Device');

    FDefaultDevice := GetDeviceInfo(Device);

  // Get the session manager for the endpoint device.
    HR := Device.Activate(IID_IAudioSessionManager, CLSCTX_INPROC_SERVER, nil,
                               IUnknown(SessionManager));
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get AudioSessionManager Interface');

  // get SimpleAudioVolume
    HR := SessionManager.GetSimpleAudioVolume(nil, 0, SimpleAudioVolume);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get SimpleAudioVolume Interface');

  // Get the control interface for the process-specific audio session.
    HR := SessionManager.GetAudioSessionControl(nil, false, SessionControl);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get AudioSessionControl');
  //  HR := SessionControl.SetGroupingParam(@mylpcguid, nil);

  // Register IAudioSessionEvents interface to receive notifications when session events occur.
    HR := SessionControl.RegisterAudioSessionNotification(SessionEvents);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to Register Session Notification Event');

  // We can get session enumerator if we use Windows 7 or newer system.
    if (Win32MajorVersion > 6) or ((Win32MajorVersion = 6) and (Win32MinorVersion >= 1)) then
    begin
    // Get the session manager2 for the endpoint device.
      HR := Device.Activate(IID_IAudioSessionManager2, CLSCTX_ALL, nil, IUnknown(SessionManager2));
      if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get AudioSessionManager2 Interface');

    // Get the session enumerator
      HR := SessionManager2.GetSessionEnumerator(SessionEnumerator);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get SessionEnumerator Interface');
    end;

    Result := true;
  finally

  end;
end;

function TAudioVolume.RefreshDeviceList(Device_Enumerator: IMMDeviceEnumerator): boolean;
var
  HR: HRESULT;
  N: Cardinal;

begin
  result := false;

  if DeviceCollection <> nil then
    DeviceCollection := nil;

  try
  // Get the device info of audio endpoint devices.
    HR := Device_Enumerator.EnumAudioEndpoints(eRender, DEVICE_STATE_ACTIVE, DeviceCollection);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to retrieve device collection');
    HR := DeviceCollection.GetCount(FDeviceCount);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get device collection length');

    if FDeviceCount > 0 then
    begin
      SetLength(FDeviceList, FDeviceCount);
      for N := 0 to (FDeviceCount - 1) do
        FDeviceList[N] := GetDeviceInfo2(DeviceCollection, N);
    end else
      if Assigned(OnError) then FOnError('Error : No device to be controlled');

    result := true;
  finally

  end;
end;

constructor TAudioVolume.Create(AOwner: TObject);
var
  HR: HRESULT;
  N: Cardinal;
  ISessionEvents: TAudioSessionEvents;
  IAudioEndpointEvents: TAudioEndpointEvents;

begin
  inherited Create;

  DeviceEnumerator := nil;
  try
  // Use Core Audio API if the operating system is Windows Vista or newer.
  // CoInitializeEx(nil, COINIT_APARTMENTTHREADED);  // COINIT_MULTITHREADED
  //  HR := CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
  //  if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to initialize Component Object Model');

    ISessionEvents := TAudioSessionEvents.Create;
    IAudioEndpointEvents := TAudioEndpointEvents.Create;
    //if not (csDesigning in ComponentState) then  !!!
      FEventHandle := AllocateHWnd(ProcessMsg);
    ISessionEvents.SetMsgHandle(FEventHandle);
    SessionEvents := ISessionEvents;
    IAudioEndpointEvents.SetMsgHandle(FEventHandle);
    AudioEndpointEvents := IAudioEndpointEvents;

  // Get the enumerator for the audio endpoint devices on this system.
    HR := CoCreateInstance(CLASS_IMMDeviceEnumerator, nil, CLSCTX_INPROC_SERVER,
              IID_IMMDeviceEnumerator, DeviceEnumerator);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to instantiate device enumerator');

  // Get the names of audio endpoint devices.
    HR := DeviceEnumerator.EnumAudioEndpoints(eRender, DEVICE_STATE_ACTIVE, DeviceCollection);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to retrieve device collection');
    HR := DeviceCollection.GetCount(FDeviceCount);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get device collection length');

    if FDeviceCount > 0 then
    begin
      SetLength(FDeviceList, FDeviceCount);
      for N := 0 to (FDeviceCount - 1) do
      //  FDeviceNames[N] := GetDeviceName(DeviceCollection, N);
        FDeviceList[N] := GetDeviceInfo2(DeviceCollection, N);
     // SelectAudioDevice(FDeviceCount - 1);  // * for device selection at creation.
    end else
      if Assigned(OnError) then FOnError('Error : No device to be controlled');

    if SetupDefaultAudioEndpoint(DeviceEnumerator) then
      FReady := true;

    MMNotificationClient := TMMNotificationClient.Create;
    MMNotificationClient.SetMsgHandle(FEventHandle);
    HR := DeviceEnumerator.RegisterEndpointNotificationCallback(MMNotificationClient);

    if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to register EndpointNotificationCallback');
  finally

  end;
end;

destructor TAudioVolume.Destroy;
begin
  if MMNotificationClient <> nil then
  begin
    DeviceEnumerator.UnregisterEndpointNotificationCallback(MMNotificationClient);
    MMNotificationClient.Free;
  end;

  if AudioEndpointVolume <> nil then
    AudioEndpointVolume.UnregisterControlChangeNotify(AudioEndpointEvents);

  if SessionControl <> nil then
    SessionControl.UnregisterAudioSessionNotification(SessionEvents);
  inherited;
end;

function TAudioVolume.SelectAudioDevice(DeviceIndex: Cardinal): Boolean;
var
  HR: HRESULT;
  Device: IMMDevice;
  pDeviceId: PWideChar;
  AudioClient: IAudioClient;
  pFormat: PWAVEFORMATEX;
begin
  Result := false;

  if FDeviceCount = 0 then
    exit;
  if DeviceIndex > (FDeviceCount - 1) then
    exit;

  if FDeviceList[DeviceIndex].DeviceId = FSelectedDevice.DeviceId then
    exit;

  try
    if AudioEndpointVolume <> nil then
    begin
      HR := AudioEndpointVolume.UnregisterControlChangeNotify(AudioEndpointEvents);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to unregister AudioEndpointEvents for device #' + int2Str(DeviceIndex));

      AudioEndpointVolume := nil;   // AudioEndpointVolume is released by automatic action
      MasterMeter := nil;   // MasterMeter is released by automatic action
    end;

    HR := DeviceCollection.Item(DeviceIndex, Device);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get IMMDevice interface for device #' + int2Str(DeviceIndex));

    HR := Device.Activate(IID_IAudioClient, CLSCTX_INPROC_SERVER, nil, IUnknown(AudioClient));
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get AudioClient interface for device #' + int2Str(DeviceIndex));
    HR := AudioClient.GetMixFormat(pformat);  // Get mixer format
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get MixFormat for device #' + int2Str(DeviceIndex));
    FMixFormat := pformat^;
    FChannels := FMixFormat.nChannels;
    CoTaskMemFree(pformat);

    HR := Device.Activate(IID_IAudioEndpointVolume, CLSCTX_INPROC_SERVER, nil, IUnknown(AudioEndpointVolume));
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get IAudioEndpointVolume interface for device #' + int2Str(DeviceIndex));

  // Get IAudioMeterInformation interface to use for peak detect
    HR := Device.Activate(IID_IAudioMeterInformation, CLSCTX_ALL, nil, IUnknown(MasterMeter));
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to get MasterMeter interface for device #' + int2Str(DeviceIndex));

    HR := AudioEndpointVolume.RegisterControlChangeNotify(AudioEndpointEvents);
    if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to register AudioEndpointEvents for device #' + int2Str(DeviceIndex));
    FSelectedDevice.DeviceName := GetDeviceName(DeviceCollection, DeviceIndex);
    Device.GetId(pDeviceId);
    FSelectedDevice.DeviceId := WideString(pDeviceId);
    Result := true;
  finally

  end;
end;

procedure TAudioVolume.ProcessMsg(var Msg: TMessage);
var
  PVolMuteRec: PVolMute;
  Volume: Single;
  Muted: Boolean;
  ByMyAction: Boolean;
  Device: IMMDevice;
  DeviceId: WideString;
  DeviceInfo: TDeviceInfo;
 // pw: PWideChar;
  HR: HRESULT;
  NewState: DWORD;
begin
  case Msg.Msg of
    WM_EndpointVolume: begin
        PVolMuteRec := PVolMute(Msg.WParam);
        ByMyAction := IsEqualGUID(TGUID(PGUID(Msg.LParam)^), GUID_TAudioVolume);
        if Assigned(FOnMasterVolumeEvent) then
          FOnMasterVolumeEvent(ByMyAction, PVolMuteRec^.Volume, PVolMuteRec^.Muted);
      end;
    WM_VolumeEvent: begin
        PVolMuteRec := PVolMute(Msg.WParam);
      // The result of ByMyAction is always false if we use IsEqualGUID function.
        ByMyAction := IsEqualGUID(TGUID(PGUID(Msg.LParam)^), GUID_TAudioVolume);
        if (MyVolSet = true) and (({abs}(PVolMuteRec^.Volume) - MyVolVal) < 0.0001) then
        begin
          ByMyAction := true;
          MyVolSet := false;
        end else
        if (MyMuteSet and (PVolMuteRec^.Muted = MyMuteVal)) then
        begin
          ByMyAction := true;
          MyMuteSet := false;
        end else
          ByMyAction := false;

        if Assigned(FOnVolumeEvent) then
          FOnVolumeEvent(ByMyAction, PVolMuteRec^.Volume, PVolMuteRec^.Muted);
      end;
    WM_SessionStateEvent: begin
        if Assigned(FOnSessionStateEvent) then FOnSessionStateEvent(Msg.WParam);
      end;
    WM_DeviceStateChange, WM_DeviceAdded, WM_DeviceRemoved: begin
        DeviceId := WideString(PWideChar(Msg.WParam));
        NewState := Msg.LParam;
        // Just refresh device list to simplify the logic.
        RefreshDeviceList(DeviceEnumerator);
        //  SetupDefaultAudioEndpoint(DeviceEnumerator);
        if Assigned(OnDeviceStateChange) then
        begin
          if Msg.Msg = WM_DeviceStateChange then
            FOnDeviceStateChange(DeviceId, NewState)
          else if Msg.Msg = WM_DeviceAdded then
            FOnDeviceStateChange(DeviceId, DEVICE_STATE_ADDED)
          else if Msg.Msg = WM_DeviceRemoved then
            FOnDeviceStateChange(DeviceId, DEVICE_STATE_REMOVED);
        end;
      end;
    WM_DefaultDeviceChange: begin
       // pw := PWideChar(Msg.WParam);
       // DeviceId := WideString(pw);
        HR := DeviceEnumerator.GetDefaultAudioEndpoint(eRender, eMultimedia, Device);
        if HR <> S_OK then if Assigned(OnError) then FOnError('Error : Unable to get DefaultAudioEndpoint Device');
        DeviceInfo := GetDeviceInfo(Device);
        if DeviceInfo.DeviceId <> FDefaultDevice.DeviceId then
        begin
          FDefaultDevice := DeviceInfo;
          if Assigned(OnDefaultDeviceChange) then
            FOnDefaultDeviceChange(DeviceInfo);
        end;
      end;
  end;
end;

procedure TAudioVolume.SetSuppressEvent(Value: Boolean);
var
  HR: HRESULT;
begin
  if Value = FSuppressEvent then Exit;
  if (SessionControl = nil) and (AudioEndpointVolume = nil) then Exit;

  if Win32MajorVersion < 6 then
  begin
    FSuppressEvent := Value;
    exit;
  end;

  try
    if AudioEndpointVolume <> nil then
    begin
      if Value then
        HR := AudioEndpointVolume.UnregisterControlChangeNotify(AudioEndpointEvents)
      else
        HR := AudioEndpointVolume.RegisterControlChangeNotify(AudioEndpointEvents);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to (un)register AudioEndpointEvents');
    end;

    if (SessionControl <> nil) then
    begin
      if Value then
        HR := SessionControl.UnregisterAudioSessionNotification(SessionEvents)
      else
        HR := SessionControl.RegisterAudioSessionNotification(SessionEvents);
      if HR <> S_OK then if Assigned(OnError) then FOnError('Unable to (un)register AudioSessionNotification');
    end;

  finally

  end;

  FSuppressEvent := Value;
end;

function TAudioVolume.AudioDeviceInfo(DeviceIndex: Cardinal): TDeviceInfo;
begin
  Result.DeviceName := '';
  Result.DeviceId := '';

  if FDeviceCount = 0 then
    exit;
  if DeviceIndex > (FDeviceCount - 1) then
    exit;

  Result := FDeviceList[DeviceIndex];
end;

function TAudioVolume.IsMasterMuted: Boolean;
var
  b: BOOL;
  HR: HRESULT;
begin
  result := false;
  if AudioEndpointVolume = nil then exit;

  HR := AudioEndpointVolume.GetMute(b);
  if HR = S_OK then
    result := b;
end;

procedure TAudioVolume.SetMasterMute(Value: Boolean);
var
  b: BOOL;
  i: integer;
  HR: HRESULT;
begin
  if AudioEndpointVolume = nil then exit;

  HR := AudioEndpointVolume.GetMute(b);
  if HR = S_OK then
    if Value <> b then
    begin
  // ** We should use very tricky code as follows.
  // ** If we use normal sentence like "AudioEndpointVolume.SetMute(Value, nil);"
  // ** it does not operate.
      if b then i := 0 else i := 1;
      AudioEndpointVolume.SetMute(BOOL(i), @GUID_TAudioVolume);
    end;
end;

function TAudioVolume.GetMasterVolume: Single;
var
  f: Single;
  HR: HRESULT;
begin
  result := 0;
  if AudioEndpointVolume = nil then exit;

   HR := AudioEndpointVolume.GetMasterVolumeLevelScalar(f);
   if HR = S_OK then
     result := f;
end;

procedure TAudioVolume.SetMasterVolume(Value: Single);
var
  I: integer;
  dwVolume: DWORD;
begin
  if Value < 0 then exit;
  if Value > 1.0 then exit;
  if AudioEndpointVolume = nil then exit;
  AudioEndpointVolume.SetMasterVolumeLevelScalar(Value, @GUID_TAudioVolume);
end;

function TAudioVolume.IsMuted: Boolean;
var
  b: LongBool;
  HR: HRESULT;
begin
  result := false;

  if Win32MajorVersion < 6 then exit;
  if SimpleAudioVolume = nil then exit;

  HR := SimpleAudioVolume.GetMute(b);
  if HR = S_OK then
    result := b;
end;


procedure TAudioVolume.SetMute(Value: Boolean);
var
  p: pointer;
begin
  if Win32MajorVersion < 6 then exit;
  if SimpleAudioVolume = nil then exit;

  p := @GUID_TAudioVolume;
  MyMuteSet := true; MyMuteVal := Value;
  SimpleAudioVolume.SetMute(Value, p);
end;

function TAudioVolume.GetVolume: Single;
var
  f: Single;
  HR: HRESULT;
begin
  result := 0;
  if Win32MajorVersion < 6 then exit;
  if SimpleAudioVolume = nil then exit;

   HR := SimpleAudioVolume.GetMasterVolume(f);
   if HR = S_OK then
     result := f;
end;

procedure TAudioVolume.SetVolume(Value: Single);
begin
  if Value < 0 then exit;
  if Value > 1.0 then exit;
  if Win32MajorVersion < 6 then exit;
  if SimpleAudioVolume = nil then exit;

  MyVolSet := true; MyVolVal := Value;
  SimpleAudioVolume.SetMasterVolume(Value, @GUID_TAudioVolume);
end;


function TAudioVolume.GetMasterPeak: Single;
var
  HR: HRESULT;
  fPeak: Single;
begin
  result := 0;

  if MasterMeter = nil then exit;

  try
    HR := MasterMeter.GetPeakValue(fPeak);
    if HR = S_OK then
      result := fPeak;
  except

  end;

end;


function TAudioVolume.GetPeak: Single;
var
  HR: HRESULT;
  fPeak: Single;
begin
  result := 0;

  if PeakMeter = nil then exit;

  try
    HR := PeakMeter.GetPeakValue(fPeak);
    if Succeeded(HR) then
      result := fPeak;
  except

  end;

end;

//-----------------------------------------------------------------------------------

procedure InitPlatformId;
var
  OSVersionInfo: TOSVersionInfo;
begin
  OSVersionInfo.dwOSVersionInfoSize := SizeOf(OSVersionInfo);
  if GetVersionEx(OSVersionInfo) then
    with OSVersionInfo do
    begin
      Win32Platform := dwPlatformId;
      Win32MajorVersion := dwMajorVersion;
      Win32MinorVersion := dwMinorVersion;
      Win32BuildNumber := dwBuildNumber;
      Win32CSDVersion := szCSDVersion;
    end;
end;

initialization
  InitPlatformId;
  if Win32MajorVersion < 6 then
  begin
    _debug(' VolumeControl    Windows XP!');
    halt;
  //Exit; // for Windows XP
  end;

  CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
finalization
  CoUninitialize();
end.
